#!/usr/bin/env python
# -*- encoding: utf-8 -*-
'''
@File    :   __init__.py
@Desc    :   Import part and configuration part.
'''

# here put the import lib


from spartan.util import *
from spartan.tensor import TensorData
